(function (strWinTypes, strAppTitles, strMaxWins, strTB, strLR, strVA,
	strClock, strDoResize, strEdgePropn, strAreaPropn) {
	'use strict';

	// OPTIONS
	var dctOpt = {
		resizeRetries: 1, // if a window is not yet resized
		resizeTolerance: 23 // margin of error for h or w size mismatch
	};

	// Set of booleans from parsed arguments
	// strTopBottom --> strLeftRight --> strVertiOrAcross --> strClockwise -->
	//  {blnTop: blnBottom: blnL: blnR: blnVert: blnHoriz: blnAntiClockwise:}
	function optionState(strTB, strLR, strVA, strClock) {
		return {
			t: !strTB.indexOf('top'),
			b: (strTB.indexOf('bottom') !== -1),
			l: !strLR.indexOf('left'),
			r: (strLR.indexOf('right') !== -1),
			v: !strVA.indexOf('vert'),
			h: (strVA.indexOf('across') !== -1),
			a: !strClock.indexOf('anti')
		};
	}

	// Full cycle of 8 tiling configurations [4 corners * (horiz | vert)]
	// filtered down to an array of [1 .. 8] options for custom cycle
	// [{lngCorner: strDirn: (blnL: | blnR:) (blnT: | blnB:) (blnH: | blnV:)}]
	// dctParams --> [{cnr: drn: b|t: r|l: h|v:}]
	function cycleChoice(a) {
		return [{
			cnr: 0,
			drn: 'down',
			l: true,
			t: true,
			v: true
    }, {
			cnr: 0,
			drn: 'right',
			l: true,
			t: true,
			h: true
    }, {
			cnr: 1,
			drn: 'left',
			r: true,
			t: true,
			h: true
    }, {
			cnr: 1,
			drn: 'down',
			r: true,
			t: true,
			v: true
    }, {
			cnr: 2,
			drn: 'up',
			r: true,
			b: true,
			v: true
    }, {
			cnr: 2,
			drn: 'left',
			r: true,
			b: true,
			h: true
    }, {
			cnr: 3,
			drn: 'right',
			l: true,
			b: true,
			h: true
    }, {
			cnr: 3,
			drn: 'up',
			l: true,
			b: true,
			v: true
    }].filter(function (b) {
			return (!b.l === !a.l || !b.r === !a.r) &&
				(!b.t === !a.t || !b.b === !a.b) &&
				(!b.h === !a.h || !b.v === !a.v);
		});
	}

	// Successor (first succedes last)
	// [c] --> i --> c
	function cycleNext(lst, i) {
		return (i < 0 || i >= (lst.length - 1)) ? 0 : i + 1;
	}

	// Predecessor (last precedes first)
	// [c] --> i --> c
	function cyclePrev(lst, i) {
		var lngLast = lst.length - 1;
		return (i < 1 || i > (lngLast + 1)) ? lngLast : i - 1;
	}

	/**************** SET OF WINDOWS ? ******************/

	// All open wins of each specified app (possibly capped)
	// or any front win of each app (possibly capped)
	// [app] --> strOpt --> maybe int --> [win]
	function winSet(lstApps, strOpt, mbMaxWins) {
		var iMax = (!mbMaxWins || isNaN(mbMaxWins)) ?
			Infinity : Number(mbMaxWins),
			blnFront = (strOpt.indexOf('front windows') === 0),
			lstWins;

		return lstApps.reduce(
			function (lstAccum, app) {
				return (
					(lstAccum.length < iMax) &&
					((lstWins = app.windows()).length) ?
					lstAccum.concat(
						blnFront ? lstWins[0] : lstWins
					) : lstAccum
				);
			}, []
		).slice(0, iMax);
	}

	// Running applications which match the options
	// strOpt --> maybe strTitles --> [app]
	function appSet(strAppTypes, mbstrTitles) {
		'use strict';

		var strTitles = (mbstrTitles || '').trim(),
			fnApps = appSE.applicationProcesses.whose(
				'KM palettes' === strAppTypes ? {
					name: 'Keyboard Maestro Engine'
				} :
				'front application' === strAppTypes ? {
					frontmost: true
				} :
				'application(s) named below' === strAppTypes && strTitles ?
				appNameFilter(
					maybeList(
						strTitles, [',', '|', ':']
					)
				) : {
					backgroundOnly: false
				}
			);

		return frontFirst(
			Object.keys(fnApps).map(function (x) {
				return fnApps[x];
			})
		);
	}
	
	// List of trimmed tokens as split by a list of potential delimiters
	// str --> [str] --> [str]
	function maybeList(strMaybeList, lstPossibleDelims) {

		return strMaybeList.split(
			new RegExp(
				'\\s*[' +
				lstPossibleDelims.reduce(function (s, d) {
					return s + d;
				}, '') +
				']\\s*'
			)
		);
	}

	// Delimited string of application names/titles --> JXA .whose({}) clause
	// (filtering application processes by the name or title string in their windows)
	// [name] --> {_or: [{(name|title):str}]}
	function appNameFilter(lstNames) {
		var strName = lstNames.length ? lstNames[0] : '';

		return (lstNames.length > 1) ? {
			_or: lstNames.reduce(
				function (lst, x) {
					lst.push({
						name: x
					});
					lst.push({
						title: x
					});
					return lst;
				}, []
			)
		} : {
			_or: [{
				name: strName
			}, {
				title: strName
			}]
		};
	}

	// List of apps with focal app at [0]
	// [app] --> [app]
	function frontFirst(lstApps) {
		var appFront = [],
			blnFront;

		return appFront.concat(
			lstApps.filter(function (app) {
				blnFront = app.frontmost();
				if (blnFront) appFront.push(app);
				return !blnFront;
			})
		);
	}

	/******************* SCREEN ? ***********************/



	// Geometric centre of window
	// ( Retiling uses the screen containing the anchor window's centre )
	// oWin --> (x, y)
	function centrePoint(oWin) {
		var lstPosn = oWin.position(),
			lstSize = oWin.size();

		return [
			lstPosn[0] + lstSize[0] / 2,
			lstPosn[1] + lstSize[1] / 2
		]
	}

	// Frame dictionary for any screen containing given point
	// [x, y] --> [{x y w h}] --> maybe {x y w h}
	function pointFrame(lstXY, lstScreens) {
		var pX = lstXY[0],
			pY = lstXY[1],
			lstContaining = (lstScreens.length > 1) ?
			lstScreens.filter(
				function (dct) {
					var sX = dct.x,
						sY = dct.y;

					return (
						(pX >= sX) && (pX < (sX + dct.w)) &&
						(pY >= sY) && (pY < (sY + dct.h))
					);
				}
			) : lstScreens;

		return lstContaining.length ? lstContaining[0] : lstScreens[0];
	}


	// Array of screen frame dictionaries
	// [strDimns] --> [{x y w h}]
	function screenArray(lstDimns) {
		var lngMenuDepth = 23; // empirical magic number of pixels


		return lstDimns.map(
			function (strDimns) {
				var lst = strDimns.split(',').map(Number);

				return {
					x: lst[0],
					y: lst[1] + lngMenuDepth,
					w: lst[2],
					h: lst[3] - lngMenuDepth
				};
			}
		);
	}


	// Four corners of a screen
	// (clockwise from top left)
	// {x y w h} --> [[x,y],[x,y],[x,y],[x,y]]
	function frameCorners(dctScreen) {
		if (!dctScreen) return null;
		var lngX = dctScreen.x,
			lngY = dctScreen.y,
			lngRight = lngX + dctScreen.w - 1,
			lngBottom = lngY + dctScreen.h - 1;

		return [
      		[lngX, lngY],
      		[lngRight, lngY],
      		[lngRight, lngBottom],
      		[lngX, lngBottom]
    	];
	}

	// The top left coordinates for fitting a window into a given corner
	// [x y] --> [w h] --> iInteger --> [x y]
	function windowTopLeftForCorner(lstCornerXY, lstWinSize, iCorner) {
		var lngX = lstCornerXY[0],
			lngY = lstCornerXY[1];

		// (X - window width) in right-hand corners, iCorner==(1, 2)
		// and (Y - window height) in lower corners  iCorner==(2, 3)
		return [
			(iCorner && (iCorner !== 3)) ?
			(lngX - lstWinSize[0]) :
      		lngX, (iCorner > 1) ?
			(lngY - lstWinSize[1]) : lngY
    	];
	}

	// Default alignment enough for this corner+direction pair ?
	// (default is top or left aligned)
	// iCorner --> blnVertical --> blnDefaultAlignment
	function cornerDirectionDefaultAligned(iCorner, blnVertical) {
		//          v   h
		// 0 (tl)   1   1
		// 1 (tr)   0   1
		// 2 (br)   0   0
		// 3 (bl)   1   0
		return blnVertical ?
			(!iCorner || (iCorner === 3)) :
			(iCorner < 2);
	}

	// Position for the next tile
	// SIDE EFFECT: applies a position update to the current tile
	// lstPrevXY --> (oWin + side effect) --> lstNewXY
	function postMoveNextTilePosn(lstPosn, oWin) {
		var lstResized = blnResize ? lstSize : oWin.size(),
			lstAligned = blnAlign ? fnAlign(
				lstPosn, lstResized
			) : lstPosn,
			iRetry = dctOpt.resizeRetries,
			lngPixels = dctOpt.resizeTolerance,
			lstNewSize;

		// REPOSITIONED
		oWin.position = blnAZ ? lstAligned :
			(blnVertical ? [
        lstAligned[0],
        lstAligned[1] - lstResized[1]
      ] : [
        lstAligned[0] - lstResized[0],
        lstAligned[1]
      ]);

		// AND POSSIBLY RESIZED
		if (blnResize) {
			oWin.size = lstResized;

			// APPLICATIONS MAY NEED TO TIME TO RESPOND TO THIS
			while (iRetry--) {
				console.log(iRetry);
				lstNewSize = oWin.size();
				if (!tileSizeEqual(lstNewSize[0], lstNewSize[1], lstResized[0],
						lstResized[1], lngPixels)) {
					oWin.size = lstResized;
				} else break;
			}
		}

		// and return an adjacent position for any successor.
		return [
      lstPosn[0] + (blnVertical ?
				0 : (blnAZ ? lstResized[0] : -lstResized[0])
      ),
      lstPosn[1] + (blnVertical ?
				(blnAZ ? lstResized[1] : -lstResized[1]) : 0
      )
    ];
	}

	// Aligned position ( right | bottom )
	// lstPosn --> lstResized --> lstAlignedPosn
	function fnAlign(lstPosn, lstResized) {
		return blnVertical ? [
      lngMaxH - lstResized[0],
      lstPosn[1]
    ] : [
      lstPosn[0],
      lngMaxY - lstResized[1]
    ];
	}

	// Array of dimension strings for connected display devices
	// () --> [main,n1,n2,n3] (the 'main' screen will appear twice)
	function screenDimns() {
		return appKME.processTokens(
			Array.apply(null,
				Array(
					Number(
						appKME.calculate('SCREENCOUNT')
					) + 1 // Main screen [0], followed by all screens [1..n]
				)
			).reduce(
				function (s, x, i) {
					return s + (s.length ? '|%Screen%' : '%Screen%') + i + '%';
				}, ''
			)
		).split('|');
	}

	// Dimensions of a tile resized to fit equally sized peers into
	// some proportion of a screen edge, covering some proportion of  area
	// lngCanvasW --> lngCanvasH --> rEdgePropn --> rAreaPropn -->
	//    blnVertical --> lngTiles --> (w, h)
	function tileSize(lngW, lngH, rEdgePropn, rAreaPropn, blnVertical,
		lngTiles) {
		return blnVertical ? [
      Math.floor(lngW * rAreaPropn),
      Math.floor((lngH * rEdgePropn) / lngTiles)
    ] : [
      Math.floor((lngW * rEdgePropn) / lngTiles),
      Math.floor(lngH * rAreaPropn)
    ];
	}

	// (Possibly tolerant) equality for tileSizes
	// [lng, lng] --> [lng, lng] --> maybeInteger --> bool
	function tileSizeEqual(x, y, x1, y1, lngMargin) {
		lngMargin = lngMargin || 0;
		return (Math.abs(x1 - x) < lngMargin) &&
			(Math.abs(y1 - y) < lngMargin);
	}

	// strName --> okmVariable
	function kmVariable(strName) {
		var kmVars = appKME.variables;

		try {
			kmVars[strCycleKey].name();
		} catch (e) {
			kmVars.push(appKME.Variable({
				name: strCycleKey,
				value: -1
			}));
		}
		return kmVars[strName];
	}

	// Check that parent apps of retiled windows have been activated
	// ( fn for array.reduce() )
	// dctActivatedApps --> win --> () --> dctActivatedApps
// 		function activeParentApp(dctSofar, win) {
// 			var strTitle = win.attributes['AXParent'].value.title();
// 	
// 	 		dctSofar[strTitle] = dctSofar[strTitle] ||
// 				appSE.activate(Application(strTitle));
// 			return dctSofar;
// 		}

	/************************** MAIN *********************************/

	// WHICH WINDOWS ?

	var appSE = Application('System Events'),
		lstApps = appSet(strWinTypes, strAppTitles);
		
	var	lstWins = winSet(
			lstApps,
			strWinTypes,
			strMaxWins
		);
				
	var	lngWins = lstWins.length,
		strMsg = (lngWins ? '' : (
			'No windows found for options/spellings:\n\n' +
			((strWinTypes === 'application(s) named below') ?
				strAppTitles : strWinTypes) + '\n\n' +
			'Maximum number of windows: ' + strMaxWins
		));
		

	if (lngWins) {

		// WHERE WERE WE ?

		var appKME = Application('Keyboard Maestro Engine'),
			strCycleKey = 'tileStageIndex',
			oMemory = kmVariable(strCycleKey), // found or created
			iPrev = Number(oMemory.value());

		// WHERE NEXT ?

		var dctParams = optionState(strTB, strLR, strVA, strClock),
			lstMenu = cycleChoice(dctParams),
			iOption = dctParams.a ?
			cyclePrev(lstMenu, iPrev) :
			cycleNext(lstMenu, iPrev);

		// APPLYING WHAT TILING PATTERN ?

		var dctChosen = lstMenu[iOption],
			iCorner = dctChosen.cnr,
			blnVertical = !dctChosen.h,
			blnAlign = !cornerDirectionDefaultAligned(iCorner, blnVertical),
			strDirn = dctChosen.drn,
			blnAZ = ((strDirn === 'down') || (strDirn === 'right')),
			blnResize = ((strDoResize === '1') &&
				(strWinTypes !== 'KM palettes') &&
				(strAppTitles !== 'Keyboard Maestro Engine'));

		// WITH WHICH SCREEN, ANCHOR WINDOW, AND TILE SIZE ?

		var oWin = lstWins[0], // Most recently moved or AXRaise d
			dctCanvas = pointFrame(
				centrePoint(oWin),
				screenArray(screenDimns())
			);

		var lstPreSize = oWin.size(),
			lstSize = blnResize ?
			tileSize(
				dctCanvas.w, dctCanvas.h,
				Number(strEdgePropn), Number(strAreaPropn),
				blnVertical,
				lngWins
			) :
			lstPreSize;

		// AT WHICH ANCHOR POSITION ?

		var lstAnchorPosn = windowTopLeftForCorner(
			frameCorners(
				dctCanvas
			)[iCorner],
			lstSize,
			iCorner
		);

		if (!blnAZ) { // marked directions: upwards or to right
			if (blnVertical) lstAnchorPosn[1] += lstSize[1];
			else lstAnchorPosn[0] += lstSize[0];
		}

		// What alignment rulers for fixed size palettes ?
		var lngMaxH = blnAlign ? lngMaxH = lstAnchorPosn[0] + lstSize[0] : null,
			lngMaxY = blnAlign ? lstAnchorPosn[1] + lstSize[1] : null;

		// AND WITH WHAT SIDE-EFFECTS ?
		
		// Each window foregrounded (reverse order conserves layering)
		var i = lngWins;
		while (i--) {
			appSE.perform(
				lstWins[i].actions['AXRaise']
			);
		}
		

		// the anchor window positioned,
		oWin.position = lstAnchorPosn;

		// and the remaining windows tiled,
		lstWins.reduce(
			postMoveNextTilePosn,
			lstAnchorPosn
		);
		

		// FINALLY, IN RETROSPECT, WHAT WILL HISTORY ASK US ?

		oMemory.value = iOption.toString();

	}
	return strMsg;

})(
	'application(s) named below',
	'TextEdit',
	'',
	'top or bottom',
	'left or right',
	'vertically or across',
	'clockwise',
	'1',
	'0.75',
	'0.25'
);

	// "$KMPARAM_Windows_to_tile_and_cycle",
	// "$KMPARAM_Application_titles",
	// "$KMPARAM_Maximum_number_of_windows",
	// "$KMPARAM_Top_and_bottom_corners", 
	// "$KMPARAM_Left_and_right_corners",
	// "$KMPARAM_Tiling",
	// "$KMPARAM_Cycle_to_next_position",
	// "$KMPARAM_Resize_adjustable_windows",
	// "$KMPARAM_Edge_of_screen_used",
	// "$KMPARAM_Area_used"


